/*
 * File:   img_acts_kernel.cu
 * Author: Alex Krizhevsky
 *
 * Created on January 25, 2011, 7:44 PM
 */

#include "../../include/kernel/img_acts_kernel.cuh"

/*
 * Block size: 16x16.
 * blockIdx.x determines case in batches of 16*imgsPerThread.
 * blockIdx.y determines 4x4 image region in target image.
 *
 * threadIdx.x determines case.
 * threadIdx.y determines pixel.
 *
 * hidActs:     (numModules, numFilters, numImages) if mfi
 *              (numFilters, numModules, numImages) otherwise
 * weights:     (numModules, numColors, filterPixels, numFilters)
 * targets:     (numColors, imgPixels, numImages)
 *
 * Each block reconstructs one 4x4 pixels from 16*imgsPerThread cases.
 *
 * Number of images must be divisible by 16*imgsPerThread.
 * 16 * imgsPerThread must be divisible by 32.
 *
 * This version loads 32 cases at a time, so it gets full coalescing on that load.
 * It only loads 16 weights at a time, so those aren't fully coalesced.
 * This version conserves shared memory by loading 16 filters at a time rather than 32.
 */
template <int imgsPerThread, int numColors, bool checkFilterBounds, bool mfi>
__global__ void img_acts_16x16_load32_batched_color_kernel(const float* hidActs, const float* weights, float* targets,
                                   const int numModulesX, const int numImages, const int numFilters,
                                   const int filterSize, const int imgSize, const int paddingStart, const int moduleStride) {
    __shared__ float shWeights[numColors*16][16 + 1];
    __shared__ float shHidActs[16][16*imgsPerThread];

    const int caseIdx = blockIdx.x * 16*imgsPerThread + threadIdx.x;
    const int numRegionsX = DIVUP(imgSize, 4);
    const int blockRegionIdx = blockIdx.y;
    const int blockRegionIdxX = blockRegionIdx % numRegionsX;
    const int blockRegionIdxY = blockRegionIdx / numRegionsX;
    const int blockRegionLeft = blockRegionIdxX * 4;
    const int blockRegionTop = blockRegionIdxY * 4;
    const int pxYInRegion = threadIdx.y / 4, pxXInRegion = threadIdx.y % 4;
    const int pxY = blockRegionTop + pxYInRegion;
    const int pxX = blockRegionLeft + pxXInRegion;
    const int pxIdx = pxY * imgSize + pxX;
    const bool isPxInImg = pxY < imgSize && pxX < imgSize;
//    const uint numModules = numModulesX * numModulesX;
    const int filterPixels = filterSize * filterSize;
    const int imgPixels = imgSize * imgSize;
    const int tidx = threadIdx.y * 16 + threadIdx.x;
    const int loadY = tidx / 32, loadX = tidx % 32;
    if (mfi) {
        hidActs += blockIdx.x * 16*imgsPerThread + loadY * numImages + loadX;
    } else {
        hidActs += blockIdx.x * 16*imgsPerThread + loadY * numImages * numModulesX * numModulesX + loadX;
    }
    weights += threadIdx.x;
    targets += pxIdx * numImages + caseIdx;
//    targets += blockRegionTop * numImages * imgSize + blockRegionLeft * numImages + (loadY / 4) *  numImages * imgSize
//             + loadY % 4 * numImages + loadX;

    float prod[numColors][imgsPerThread];
    #pragma unroll
    for (int c = 0; c < numColors; c++) {
        #pragma unroll
        for (int i = 0; i < imgsPerThread; i++) {
            prod[c][i] = 0;
        }
    }
    const int startY = blockRegionTop - paddingStart < filterSize ? 0
                        : 1 + (blockRegionTop - paddingStart - filterSize) / moduleStride;
    const int endY = MIN(numModulesX, 1 + (blockRegionTop + 4 - paddingStart) / moduleStride);
    const int startX = blockRegionLeft - paddingStart < filterSize ? 0
                        : 1 + (blockRegionLeft - paddingStart - filterSize) / moduleStride;
    const int endX = MIN(numModulesX, 1 + (blockRegionLeft + 4 - paddingStart) / moduleStride);
    
    float* shWeightLoad = &shWeights[threadIdx.y][threadIdx.x];
    float* shHidActLoad = &shHidActs[loadY][loadX];

    for (int my = startY; my < endY; my++) {
        const int moduleTop = paddingStart + my * moduleStride;
        const int pxInModuleY = pxY - moduleTop;

        for (int mx = startX; mx < endX; mx++) {
            const int moduleIdx = my * numModulesX + mx;
            const int moduleLeft = paddingStart + mx * moduleStride;
            const int pxInModuleX = pxX - moduleLeft;

            const bool isPxInModule = pxInModuleY >= 0 && pxInModuleY < filterSize && pxInModuleX >= 0 && pxInModuleX < filterSize;
            const int pxIdxInModule = pxInModuleY * filterSize + pxInModuleX;

            for (int f = 0; f < numFilters; f += 16) { // multipply with 16 filters at a time
                // Now the threads split up into half-warps, and each half-warp decides if it's interested.

                if (mfi) {
                    #pragma unroll
                    for (int j = 0; j < 16; j += 8) { // load 16 rows of imgsPerThread*16 cols, 8 * 32 elements at a time.
                        if (!checkFilterBounds || f + j + loadY < numFilters) {
                            #pragma unroll
                            for (int i = 0; i < imgsPerThread * 16; i += 32) {
                                shHidActLoad[j * 16 * imgsPerThread + i] = hidActs[(moduleIdx * numFilters + f + j) * numImages + i];
                            }
                        }
                    }
                } else {
                    #pragma unroll
                    for (int j = 0; j < 16; j += 8) { // load 16 rows of imgsPerThread*16 cols, 8 * 32 elements at a time.
                        if (!checkFilterBounds || f + j + loadY < numFilters) {
                            #pragma unroll
                            for (int i = 0; i < imgsPerThread * 16; i += 32) {
                                shHidActLoad[j * 16 * imgsPerThread + i] = hidActs[(moduleIdx + (f + j) * numModulesX * numModulesX) * numImages + i];
                            }
                        }
                    }
                }

                if (isPxInImg && isPxInModule) {
                    // This half-warp is interested, so it's going to load the weights from this module to its pixel.
                    if (checkFilterBounds) {
                        #pragma unroll
                        for (int c = 0; c < numColors; c++) {
                            shWeightLoad[c * 16 * (16 + 1)] = 0;
                        }
                    }
                    // Not fully coalesced read :(
                    // But taking out this read entirely only reduces the runtime by ~2.8%, so it isn't costing me much.
                    if (!checkFilterBounds || threadIdx.x + f < numFilters) {
                        #pragma unroll
                        for (int c = 0; c < numColors; c++) {
                            shWeightLoad[c * 16 * (16 + 1)] = weights[((moduleIdx*numColors + c) * filterPixels + pxIdxInModule) * numFilters + f];
                        }
                    }
                }

                __syncthreads();
                // Do some actual computation
                if (isPxInImg && isPxInModule) {
                    #pragma unroll
                    for (int c = 0; c < numColors; c++) {
                        #pragma unroll
                        for (int w = 0; w < 16; w++) {
                            #pragma unroll
                            for (int i = 0; i < imgsPerThread; i++) {
                                prod[c][i] += shWeights[threadIdx.y + c * 16][w] * shHidActs[w][threadIdx.x + i * 16];
                            }
                        }
                    }
                }
                __syncthreads();
            }
        }
    }
    // Not fully coalesced write :(... shmem (and fully coalesced) version below is actually slightly slower, though
    if (isPxInImg) {
        #pragma unroll
        for (int c = 0; c < numColors; c++) {
            #pragma unroll
            for (int i = 0; i < imgsPerThread; i++) {
                targets[c * imgPixels * numImages + i * 16] = prod[c][i];
        //        shHidActs[threadIdx.y][threadIdx.x + i * 16] = prod[i];
            }
        }
    }
//    __syncthreads();
//    #pragma unroll
//    for (int i = 0; i < imgsPerThread/2; i++) {
//        targets[i * 32] = shHidActs[loadY][loadX + i * 32];
//        targets[2 * numImages * imgSize + i * 32] = shHidActs[loadY + 8][loadX + i * 32];
//    }
}


/*
 * Block size: 16x16.
 * blockIdx.x determines case in batches of 16*imgsPerThread, also color in batches of colorsPerThread.
 *  In essence, blockIdx.x.x = 1..numImages/(16*imgsPerThread)
 *              blockIdx.y.y = 1..numColors/colorsPerThread
 * blockIdx.y determines 4x4 image region in target image.
 *
 * threadIdx.x determines case.
 * threadIdx.y determines pixel.
 *
 * hidActs:     (numModules, numFilters, numImages) if mfi
 *              (numFilters, numModules, numImages) otherwise
 * weights:     (numModules, numColors, filterPixels, numFilters)
 * targets:     (numColors, imgPixels, numImages)
 *
 * Each block reconstructs one 4x4 pixels from 16*imgsPerThread cases.
 *
 * Number of images must be divisible by 16*imgsPerThread.
 * 16 * imgsPerThread must be divisible by 32.
 * numColors must be divisible by colorsPerThread.
 *
 * This version loads 32 cases at a time, so it gets full coalescing on that load.
 * It only loads 16 weights at a time, so those aren't fully coalesced.
 * This version conserves shared memory by loading 16 filters at a time rather than 32.
 */
template <int imgsPerThread, int colorsPerThread, bool checkFilterBounds, bool mfi>
__global__ void img_acts_16x16_load32_batched_manycolor_kernel(const float* hidActs, const float* weights, float* targets,
                                   const int numModulesX, const int numImages, const int numFilters,
                                   const int filterSize, const int imgSize, const int paddingStart, const int moduleStride,
                                   const int numColors) {
    __shared__ float shWeights[colorsPerThread*16][16 + 1];
    __shared__ float shHidActs[16][16*imgsPerThread];

    const int blockCaseIdx = (blockIdx.x % (numImages/(16*imgsPerThread))) * 16*imgsPerThread;
    const int blockColorIdx = (blockIdx.x / (numImages/(16*imgsPerThread))) * colorsPerThread;
    const int numRegionsX = DIVUP(imgSize, 4);
    const int blockRegionIdx = blockIdx.y;
    const int blockRegionIdxX = blockRegionIdx % numRegionsX;
    const int blockRegionIdxY = blockRegionIdx / numRegionsX;
    const int blockRegionLeft = blockRegionIdxX * 4;
    const int blockRegionTop = blockRegionIdxY * 4;
    const int pxYInRegion = threadIdx.y / 4, pxXInRegion = threadIdx.y % 4;
    const int pxY = blockRegionTop + pxYInRegion;
    const int pxX = blockRegionLeft + pxXInRegion;
    const int pxIdx = pxY * imgSize + pxX;
    const bool isPxInImg = pxY < imgSize && pxX < imgSize;
//    const uint numModules = numModulesX * numModulesX;
    const int filterPixels = filterSize * filterSize;
    const int imgPixels = imgSize * imgSize;
    const int tidx = threadIdx.y * 16 + threadIdx.x;
    const int loadY = tidx / 32, loadX = tidx % 32;
    if (mfi) {
        hidActs += blockCaseIdx + loadY * numImages + loadX;
    } else {
        hidActs += blockCaseIdx + loadY * numImages * numModulesX * numModulesX + loadX;
    }
    weights += blockColorIdx * filterPixels * numFilters + threadIdx.x;
    targets += blockColorIdx * imgPixels * numImages + pxIdx * numImages + blockCaseIdx + threadIdx.x;

//    targets += blockRegionTop * numImages * imgSize + blockRegionLeft * numImages + (loadY / 4) *  numImages * imgSize
//             + loadY % 4 * numImages + loadX;

    float prod[colorsPerThread][imgsPerThread];
    #pragma unroll
    for (int c = 0; c < colorsPerThread; c++) {
        #pragma unroll
        for (int i = 0; i < imgsPerThread; i++) {
            prod[c][i] = 0;
        }
    }
    const int startY = blockRegionTop - paddingStart < filterSize ? 0
                        : 1 + (blockRegionTop - paddingStart - filterSize) / moduleStride;
    const int endY = MIN(numModulesX, 1 + (blockRegionTop + 4 - paddingStart) / moduleStride);
    const int startX = blockRegionLeft - paddingStart < filterSize ? 0
                        : 1 + (blockRegionLeft - paddingStart - filterSize) / moduleStride;
    const int endX = MIN(numModulesX, 1 + (blockRegionLeft + 4 - paddingStart) / moduleStride);

    float* shWeightLoad = &shWeights[threadIdx.y][threadIdx.x];
    float* shHidActLoad = &shHidActs[loadY][loadX];

    for (int my = startY; my < endY; my++) {
        const int moduleTop = paddingStart + my * moduleStride;
        const int pxInModuleY = pxY - moduleTop;

        for (int mx = startX; mx < endX; mx++) {
            const int moduleIdx = my * numModulesX + mx;
            const int moduleLeft = paddingStart + mx * moduleStride;
            const int pxInModuleX = pxX - moduleLeft;

            const bool isPxInModule = pxInModuleY >= 0 && pxInModuleY < filterSize && pxInModuleX >= 0 && pxInModuleX < filterSize;
            const int pxIdxInModule = pxInModuleY * filterSize + pxInModuleX;

            for (int f = 0; f < numFilters; f += 16) { // multipply with 16 filters at a time
                // Now the threads split up into half-warps, and each half-warp decides if it's interested.

                if (mfi) { // shame this if statement has to be so early...
                    #pragma unroll
                    for (int j = 0; j < 16; j += 8) { // load 16 rows of imgsPerThread*16 cols, 8 * 32 elements at a time.
                        if (!checkFilterBounds || f + j + loadY < numFilters) {
                            #pragma unroll
                            for (int i = 0; i < imgsPerThread * 16; i += 32) {
                                shHidActLoad[j * 16 * imgsPerThread + i] = hidActs[(moduleIdx * numFilters + f + j) * numImages + i];
                            }
                        }
                    }
                } else {
                    #pragma unroll
                    for (int j = 0; j < 16; j += 8) { // load 16 rows of imgsPerThread*16 cols, 8 * 32 elements at a time.
                        if (!checkFilterBounds || f + j + loadY < numFilters) {
                            #pragma unroll
                            for (int i = 0; i < imgsPerThread * 16; i += 32) {
                                shHidActLoad[j * 16 * imgsPerThread + i] = hidActs[(moduleIdx + (f + j) * numModulesX * numModulesX) * numImages + i];
                            }
                        }
                    }
                }

                if (isPxInImg && isPxInModule) {
                    // This half-warp is interested, so it's going to load the weights from this module to its pixel.
                    if (checkFilterBounds) {
                        #pragma unroll
                        for (int c = 0; c < colorsPerThread; c++) {
                            shWeightLoad[c * 16 * (16 + 1)] = 0;
                        }
                    }
                    // Not fully coalesced read :(
                    // But taking out this read entirely only reduces the runtime by ~2.8%, so it isn't costing me much.
                    if (!checkFilterBounds || threadIdx.x + f < numFilters) {
                        #pragma unroll
                        for (int c = 0; c < colorsPerThread; c++) {
                            shWeightLoad[c * 16 * (16 + 1)] = weights[((moduleIdx*numColors + c) * filterPixels + pxIdxInModule) * numFilters + f];
                        }
                    }
                }

                __syncthreads();
                // Do some actual computation
                if (isPxInImg && isPxInModule) {
                    #pragma unroll
                    for (int c = 0; c < colorsPerThread; c++) {
                        #pragma unroll
                        for (int w = 0; w < 16; w++) {
                            #pragma unroll
                            for (int i = 0; i < imgsPerThread; i++) {
                                prod[c][i] += shWeights[threadIdx.y + c * 16][w] * shHidActs[w][threadIdx.x + i * 16];
                            }
                        }
                    }
                }
                __syncthreads();
            }
        }
    }
    // Not fully coalesced write :(... shmem (and fully coalesced) version below is actually slightly slower, though
    if (isPxInImg) {
        #pragma unroll
        for (int c = 0; c < colorsPerThread; c++) {
            #pragma unroll
            for (int i = 0; i < imgsPerThread; i++) {
                targets[c * imgPixels * numImages + i * 16] = prod[c][i];
        //        shHidActs[threadIdx.y][threadIdx.x + i * 16] = prod[i];
            }
        }
    }
//    __syncthreads();
//    #pragma unroll
//    for (int i = 0; i < imgsPerThread/2; i++) {
//        targets[i * 32] = shHidActs[loadY][loadX + i * 32];
//        targets[2 * numImages * imgSize + i * 32] = shHidActs[loadY + 8][loadX + i * 32];
//    }
}

/*
 * hidActs:     (numModules, numFilters, numImages) if hidActsOrder == MODULE_FILTER_IMAGE
 *              (numFilters, numModules, numImages) otherwise
 * filters:     (numModules, numColors, filterPixels, numFilters)
 * targets:     (numColors, imgPixels, numImages)
 *
 * hidActs: The filter activity matrix.
 * filters: The filters matrix.
 * targets: Result matrix.
 * imgSize: the width (or equivalently height) of the image.
 * filterSize: the width (or equivalently height) of the filter.
 * paddingStart: non-positive number indicating where the first filter should be applied.
 * moduleStride: stride between filter applications.
 * numColors: number of color channels in images and filters.
 * hidActsOrder: how the hidActs matrix is laid out (see hidActs comment above).
 */
void computeImgActs(NVMatrix& hidActs, NVMatrix& filters, NVMatrix& targets,
                    int imgSize, int paddingStart, int moduleStride, int numColors, FILTER_OUTPUT_ORDER hidActsOrder) {
    assert(numColors > 0 && (numColors <= 3 || numColors % 2 == 0));

    int numImages = hidActs.getNumCols();
    int numFilters = filters.getNumCols();
    int numModules = hidActs.getNumRows() / numFilters;
    int filterPixels = filters.getNumRows() / (numColors*numModules);
    int filterSize = sqrt(filterPixels);
    int imgPixels = imgSize * imgSize;
    int numModulesX = sqrt(numModules);
    assert(filterPixels == filterSize * filterSize);
    assert(hidActs.getNumRows() == numModules * numFilters);
    assert(filters.getNumRows() == numColors*numModules * filterPixels);
    assert(numModules == numModulesX * numModulesX);

    assert(hidActs.isContiguous());
    assert(filters.isContiguous());

    assert(!hidActs.isTrans());
    assert(!filters.isTrans());
    assert(!targets.isTrans());
    assert(numImages % 128 == 0);
//    assert(imgPixels % 16 == 0);
    targets.resize(numColors*imgPixels, numImages);
    assert(targets.isContiguous()); // no stride support here!

    dim3 threads(16,16);

    if (hidActsOrder == MODULE_FILTER_IMAGE) {
        if (numColors > 3) {
            int colorsPerThread = numColors % 4 == 0 ? 4 : 2;
            dim3 blocks((numColors / colorsPerThread) * numImages/(16*8), DIVUP(imgSize,4) * DIVUP(imgSize,4));
            if (colorsPerThread == 4) {
                if (numFilters % 32 == 0) {
                    cudaFuncSetCacheConfig(img_acts_16x16_load32_batched_manycolor_kernel<8, 4, false, true>, cudaFuncCachePreferShared);
                    img_acts_16x16_load32_batched_manycolor_kernel<8, 4, false, true><<<blocks, threads>>>(hidActs.getDevData(), filters.getDevData(), targets.getDevData(),
                                                        numModulesX, numImages, numFilters, filterSize, imgSize, paddingStart, moduleStride, numColors);
                } else {
                    cudaFuncSetCacheConfig(img_acts_16x16_load32_batched_manycolor_kernel<8, 4, true, true>, cudaFuncCachePreferShared);
                    img_acts_16x16_load32_batched_manycolor_kernel<8, 4, true, true><<<blocks, threads>>>(hidActs.getDevData(), filters.getDevData(), targets.getDevData(),
                                                        numModulesX, numImages, numFilters, filterSize, imgSize, paddingStart, moduleStride, numColors);
                }
            } else {
                if (numFilters % 32 == 0) {
                    cudaFuncSetCacheConfig(img_acts_16x16_load32_batched_manycolor_kernel<8, 2, false, true>, cudaFuncCachePreferShared);
                    img_acts_16x16_load32_batched_manycolor_kernel<8, 2, false, true><<<blocks, threads>>>(hidActs.getDevData(), filters.getDevData(), targets.getDevData(),
                                                        numModulesX, numImages, numFilters, filterSize, imgSize, paddingStart, moduleStride, numColors);
                } else {
                    cudaFuncSetCacheConfig(img_acts_16x16_load32_batched_manycolor_kernel<8, 2, true, true>, cudaFuncCachePreferShared);
                    img_acts_16x16_load32_batched_manycolor_kernel<8, 2, true, true><<<blocks, threads>>>(hidActs.getDevData(), filters.getDevData(), targets.getDevData(),
                                                        numModulesX, numImages, numFilters, filterSize, imgSize, paddingStart, moduleStride, numColors);
                }
            }
        } else {
            dim3 blocks(numImages/(16*8), DIVUP(imgSize,4) * DIVUP(imgSize,4));
            // template <int imgsPerThread, int colors, bool checkFilterBounds>
            if (numColors == 1) {
                if (numFilters % 32 == 0) {
                    cudaFuncSetCacheConfig(img_acts_16x16_load32_batched_color_kernel<8, 1, false, true>, cudaFuncCachePreferShared);
                    img_acts_16x16_load32_batched_color_kernel<8, 1, false, true><<<blocks, threads>>>(hidActs.getDevData(), filters.getDevData(), targets.getDevData(),
                                                        numModulesX, numImages, numFilters, filterSize, imgSize, paddingStart, moduleStride);
                } else {
                    cudaFuncSetCacheConfig(img_acts_16x16_load32_batched_color_kernel<8, 1, true, true>, cudaFuncCachePreferShared);
                    img_acts_16x16_load32_batched_color_kernel<8, 1, true, true><<<blocks, threads>>>(hidActs.getDevData(), filters.getDevData(), targets.getDevData(),
                                                        numModulesX, numImages, numFilters, filterSize, imgSize, paddingStart, moduleStride);
                }
            } else if (numColors == 2) {
                if (numFilters % 32 == 0) {
                    cudaFuncSetCacheConfig(img_acts_16x16_load32_batched_color_kernel<8, 2, false, true>, cudaFuncCachePreferShared);
                    img_acts_16x16_load32_batched_color_kernel<8, 2, false, true><<<blocks, threads>>>(hidActs.getDevData(), filters.getDevData(), targets.getDevData(),
                                                        numModulesX, numImages, numFilters, filterSize, imgSize, paddingStart, moduleStride);
                } else {
                    cudaFuncSetCacheConfig(img_acts_16x16_load32_batched_color_kernel<8, 2, true, true>, cudaFuncCachePreferShared);
                    img_acts_16x16_load32_batched_color_kernel<8, 2, true, true><<<blocks, threads>>>(hidActs.getDevData(), filters.getDevData(), targets.getDevData(),
                                                        numModulesX, numImages, numFilters, filterSize, imgSize, paddingStart, moduleStride);
                }
            } else if (numColors == 3) {
                if (numFilters % 32 == 0) {
                    cudaFuncSetCacheConfig(img_acts_16x16_load32_batched_color_kernel<8, 3, false, true>, cudaFuncCachePreferShared);
                    img_acts_16x16_load32_batched_color_kernel<8, 3, false, true><<<blocks, threads>>>(hidActs.getDevData(), filters.getDevData(), targets.getDevData(),
                                                        numModulesX, numImages, numFilters, filterSize, imgSize, paddingStart, moduleStride);
                } else {
                    cudaFuncSetCacheConfig(img_acts_16x16_load32_batched_color_kernel<8, 3, true, true>, cudaFuncCachePreferShared);
                    img_acts_16x16_load32_batched_color_kernel<8, 3, true, true><<<blocks, threads>>>(hidActs.getDevData(), filters.getDevData(), targets.getDevData(),
                                                        numModulesX, numImages, numFilters, filterSize, imgSize, paddingStart, moduleStride);
                }
            }
        }
    } else {
        if (numColors > 3) {
            int colorsPerThread = numColors % 4 == 0 ? 4 : 2;
            dim3 blocks((numColors / colorsPerThread) * numImages/(16*8), DIVUP(imgSize,4) * DIVUP(imgSize,4));
            if (colorsPerThread == 4) {
                if (numFilters % 32 == 0) {
                    cudaFuncSetCacheConfig(img_acts_16x16_load32_batched_manycolor_kernel<8, 4, false, false>, cudaFuncCachePreferShared);
                    img_acts_16x16_load32_batched_manycolor_kernel<8, 4, false, false><<<blocks, threads>>>(hidActs.getDevData(), filters.getDevData(), targets.getDevData(),
                                                        numModulesX, numImages, numFilters, filterSize, imgSize, paddingStart, moduleStride, numColors);
                } else {
                    cudaFuncSetCacheConfig(img_acts_16x16_load32_batched_manycolor_kernel<8, 4, true, false>, cudaFuncCachePreferShared);
                    img_acts_16x16_load32_batched_manycolor_kernel<8, 4, true, false><<<blocks, threads>>>(hidActs.getDevData(), filters.getDevData(), targets.getDevData(),
                                                        numModulesX, numImages, numFilters, filterSize, imgSize, paddingStart, moduleStride, numColors);
                }
            } else {
                if (numFilters % 32 == 0) {
                    cudaFuncSetCacheConfig(img_acts_16x16_load32_batched_manycolor_kernel<8, 2, false, false>, cudaFuncCachePreferShared);
                    img_acts_16x16_load32_batched_manycolor_kernel<8, 2, false, false><<<blocks, threads>>>(hidActs.getDevData(), filters.getDevData(), targets.getDevData(),
                                                        numModulesX, numImages, numFilters, filterSize, imgSize, paddingStart, moduleStride, numColors);
                } else {
                    cudaFuncSetCacheConfig(img_acts_16x16_load32_batched_manycolor_kernel<8, 2, true, false>, cudaFuncCachePreferShared);
                    img_acts_16x16_load32_batched_manycolor_kernel<8, 2, true, false><<<blocks, threads>>>(hidActs.getDevData(), filters.getDevData(), targets.getDevData(),
                                                        numModulesX, numImages, numFilters, filterSize, imgSize, paddingStart, moduleStride, numColors);
                }
            }
        } else {
            dim3 blocks(numImages/(16*8), DIVUP(imgSize,4) * DIVUP(imgSize,4));
            // template <int imgsPerThread, int colors, bool checkFilterBounds>
            if (numColors == 1) {
                if (numFilters % 32 == 0) {
                    cudaFuncSetCacheConfig(img_acts_16x16_load32_batched_color_kernel<8, 1, false, false>, cudaFuncCachePreferShared);
                    img_acts_16x16_load32_batched_color_kernel<8, 1, false, false><<<blocks, threads>>>(hidActs.getDevData(), filters.getDevData(), targets.getDevData(),
                                                        numModulesX, numImages, numFilters, filterSize, imgSize, paddingStart, moduleStride);
                } else {
                    cudaFuncSetCacheConfig(img_acts_16x16_load32_batched_color_kernel<8, 1, true, false>, cudaFuncCachePreferShared);
                    img_acts_16x16_load32_batched_color_kernel<8, 1, true, false><<<blocks, threads>>>(hidActs.getDevData(), filters.getDevData(), targets.getDevData(),
                                                        numModulesX, numImages, numFilters, filterSize, imgSize, paddingStart, moduleStride);
                }
            } else if (numColors == 2) {
                if (numFilters % 32 == 0) {
                    cudaFuncSetCacheConfig(img_acts_16x16_load32_batched_color_kernel<8, 2, false, false>, cudaFuncCachePreferShared);
                    img_acts_16x16_load32_batched_color_kernel<8, 2, false, false><<<blocks, threads>>>(hidActs.getDevData(), filters.getDevData(), targets.getDevData(),
                                                        numModulesX, numImages, numFilters, filterSize, imgSize, paddingStart, moduleStride);
                } else {
                    cudaFuncSetCacheConfig(img_acts_16x16_load32_batched_color_kernel<8, 2, true, false>, cudaFuncCachePreferShared);
                    img_acts_16x16_load32_batched_color_kernel<8, 2, true, false><<<blocks, threads>>>(hidActs.getDevData(), filters.getDevData(), targets.getDevData(),
                                                        numModulesX, numImages, numFilters, filterSize, imgSize, paddingStart, moduleStride);
                }
            } else if (numColors == 3) {
                if (numFilters % 32 == 0) {
                    cudaFuncSetCacheConfig(img_acts_16x16_load32_batched_color_kernel<8, 3, false, false>, cudaFuncCachePreferShared);
                    img_acts_16x16_load32_batched_color_kernel<8, 3, false, false><<<blocks, threads>>>(hidActs.getDevData(), filters.getDevData(), targets.getDevData(),
                                                        numModulesX, numImages, numFilters, filterSize, imgSize, paddingStart, moduleStride);
                } else {
                    cudaFuncSetCacheConfig(img_acts_16x16_load32_batched_color_kernel<8, 3, true, false>, cudaFuncCachePreferShared);
                    img_acts_16x16_load32_batched_color_kernel<8, 3, true, false><<<blocks, threads>>>(hidActs.getDevData(), filters.getDevData(), targets.getDevData(),
                                                        numModulesX, numImages, numFilters, filterSize, imgSize, paddingStart, moduleStride);
                }
            }
        }
    }

    cutilCheckMsg("computeImgActs: kernel execution failed");
}